from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional

from janis_core.types import Filename, String

from janis_core.tool.tool import Tool, TOutput, TInput, ToolType, ToolTypes


class CodeTool(Tool, ABC):

    # User should inherit from these blocks

    @abstractmethod
    def inputs(self) -> List[TInput]:
        pass

    @abstractmethod
    def outputs(self) -> List[TOutput]:
        pass

    def memory(self, hints: Dict[str, Any]) -> Optional[float]:
        """
        These values are used to generate a separate runtime.json / runtime.yaml input
        that can be passed to the execution engine to fill in for the specified hints.

        These are now (2019-04-10) to be kept out of the workflow, to leave the workflow
        truly portable.

        This memory must be in GB!
        :param hints: Dict[Key: value] of hints
        :return: Optional[int]
        """
        return None

    def cpus(self, hints: Dict[str, Any]) -> Optional[int]:
        """
        These values are used to generate a separate runtime.json / runtime.yaml input
        that can be passed to the execution engine to fill in for the specified hints.

        These are now (2019-04-10) to be kept out of the workflow, to leave the workflow
        truly portable.

        The CPU must be a whole number. If your tool contains threads
        :return:
        """
        return None

    # Janis developer should inherit these methods

    @abstractmethod
    def base_command(self):
        pass

    @abstractmethod
    def script_name(self):
        pass

    @abstractmethod
    def container(self):
        pass

    @abstractmethod
    def prepared_script(self):
        pass

    # Other internal methods

    @classmethod
    def type(cls) -> ToolType:
        return ToolTypes.CodeTool

    def tool_inputs(self) -> List[TInput]:
        return self.inputs()

    def tool_outputs(self) -> List[TOutput]:
        return self.outputs()

    def generate_inputs_override(
        self,
        additional_inputs=None,
        with_resource_overrides=False,
        hints=None,
        include_defaults=True,
    ):
        d, ad = {}, additional_inputs or {}
        for i in self.inputs():
            if (
                not i.intype.optional
                or i.id() in ad
                or (include_defaults and i.default)
            ):
                d[i] = ad.get(i.id(), i.default)

        if with_resource_overrides:
            cpus = self.cpus(hints) or 1
            mem = self.memory(hints)
            d.update(
                {
                    "runtime_memory": mem,
                    "runtime_cpu": cpus,
                    "runtime_disks": "local-disk 60 SSD",
                }
            )

    def translate(
        self,
        translation: str,
        to_console=True,
        to_disk=False,
        export_path=None,
        with_docker=True,
        with_resource_overrides=False,
    ):
        from janis_core import translations

        return translations.translate_code_tool(
            self,
            translation=translation,
            to_console=to_console,
            to_disk=to_disk,
            with_docker=with_docker,
            export_path=export_path,
        )

    def wrapped_in_wf(self):
        from copy import copy
        from janis_core.workflow.workflow import WorkflowBuilder

        wf = WorkflowBuilder(self.id() + "Wf")
        inpmap = {}
        for i in self.inputs():

            if isinstance(i.intype, Filename):
                intp = String(optional=True)
            else:
                intp = copy(i.intype)
                if i.default:
                    intp.optional = True

            inpmap[i.id()] = wf.input(i.id(), intp)

        stp = wf.step(self.id().lower(), self(**inpmap))

        for o in self.outputs():
            wf.output(o.id(), source=stp[o.id()])

        return wf



import typing


is_py38 = hasattr(typing, "get_args")
is_py37 = not (is_py38) and hasattr(typing, "_GenericAlias")
is_py36 = not (is_py38 or is_py37) and hasattr(typing, "_Union")
is_py35 = not (is_py38 or is_py37 or is_py36)

if is_py38:

    def _is_generic(cls):
        if isinstance(cls, typing._GenericAlias):
            return True

        if isinstance(cls, typing._SpecialForm):
            return cls not in {typing.Any}

        return False

    def _is_base_generic(cls):
        args = typing.get_args(cls)
        args_are_empty = len(list(a for a in args if a)) == 0
        return args_are_empty or str(args[0]) == "~T"


elif is_py37:
    # python 3.7
    def _is_generic(cls):
        if isinstance(cls, typing._GenericAlias):
            return True

        if isinstance(cls, typing._SpecialForm):
            return cls not in {typing.Any}

        return False

    def _is_base_generic(cls):
        if isinstance(cls, typing._GenericAlias):
            if cls.__origin__ in {typing.Generic, typing._Protocol}:
                return False

            if isinstance(cls, typing._VariadicGenericAlias):
                return True

            return len(cls.__parameters__) > 0

        if isinstance(cls, typing._SpecialForm):
            return cls._name in {"ClassVar", "Union", "Optional"}

        return False


elif is_py36:
    # python 3.6
    def _is_generic(cls):
        return isinstance(
            cls, (typing.GenericMeta, typing._Union, typing._Optional, typing._ClassVar)
        )

    def _is_base_generic(cls):
        if isinstance(cls, (typing.GenericMeta, typing._Union)):
            return cls.__args__ in {None, ()}

        return isinstance(cls, typing._Optional)


else:
    # python 3.5
    def _is_generic(cls):
        return isinstance(
            cls,
            (
                typing.GenericMeta,
                typing.UnionMeta,
                typing.OptionalMeta,
                typing.CallableMeta,
                typing.TupleMeta,
            ),
        )

    def _is_base_generic(cls):
        if isinstance(cls, typing.GenericMeta):
            return all(isinstance(arg, typing.TypeVar) for arg in cls.__parameters__)

        if isinstance(cls, typing.UnionMeta):
            return cls.__union_params__ is None

        if isinstance(cls, typing.TupleMeta):
            return cls.__tuple_params__ is None

        if isinstance(cls, typing.CallableMeta):
            return cls.__args__ is None

        if isinstance(cls, typing.OptionalMeta):
            return True

        return False


def is_generic(cls):
    """
    Detects any kind of generic, for example `List` or `List[int]`. This includes "special" types like
    Union and Tuple - anything that's subscriptable, basically.
    """
    return _is_generic(cls)


def is_base_generic(cls):
    """
    Detects generic base classes, for example `List` (but not `List[int]`)
    """
    return _is_base_generic(cls)


def is_qualified_generic(cls):
    """
    Detects generics with arguments, for example `List[int]` (but not `List`)
    """
    return is_generic(cls) and not is_base_generic(cls)



from enum import Enum
from typing import Optional


class DocumentationMeta:
    """
    Base class to provide doc tags with more information tags could be a
    """

    def __init__(self, doc: Optional[str]):
        self.doc = doc

    def __str__(self):
        return self.doc or ""


class InputQualityType(Enum):
    user = "user"
    static = "static"
    configuration = "configuration"


class InputDocumentation(DocumentationMeta):
    def __init__(
        self,
        doc: Optional[str],
        quality: InputQualityType = None,
        example: Optional[str] = None,
    ):
        super().__init__(doc)

        self.quality = quality
        self.example = example


class OutputDocumentation(DocumentationMeta):
    def __init__(self, doc: Optional[str]):
        super().__init__(doc)



from abc import ABC, abstractmethod
from textwrap import dedent
from typing import Dict, Any, Optional, Type, List

from janis_core.tool.documentation import InputDocumentation
from janis_core.types.data_types import NativeTypes

from janis_core.utils.docparser_info import parse_docstring
from janis_core.utils.metadata import ToolMetadata

from janis_core.code.codetool import CodeTool
from janis_core.tool.tool import TInput
from janis_core.types import (
    Boolean,
    Array,
    Int,
    Float,
    String,
    DataType,
    File,
    Directory,
    get_instantiated_type,
    all_types,
)

inspect_ignore_keys = {"self", "args", "kwargs", "cls", "template"}


class PythonTool(CodeTool, ABC):
    @staticmethod
    @abstractmethod
    def code_block(**kwargs) -> Dict[str, Any]:
        """
        This code block must be 100% self contained. All libraries and functions must be
        imported and declared from within this block.
        :param kwargs:
        :return:
        """
        pass

    def __init__(self, **connections):
        self._cached_input_signature = None
        super().__init__(metadata_class=ToolMetadata, **connections)

    # Other internal methods

    def inputs(self):
        if self._cached_input_signature is None:

            import inspect

            argspec = inspect.signature(self.code_block)
            docstrings = parse_docstring(self.code_block.__doc__)
            paramlist = docstrings.get("params", [])
            paramdocs = {
                p["name"]: p.get("doc").strip() for p in paramlist if "name" in p
            }

            missing_annotations = set()
            unsupported_types = {}

            ins = []
            for inp in argspec.parameters.values():
                if inp.name in inspect_ignore_keys:
                    continue
                fdefault = inp.default
                optional = (fdefault is not inspect.Parameter.empty) or fdefault is None
                default = fdefault if optional else None

                defaulttype = type(fdefault) if fdefault is not None else None
                annotation = (
                    defaulttype
                    if inp.annotation is inspect.Parameter.empty
                    else inp.annotation
                )
                if not annotation:
                    missing_annotations.add(inp.name)
                    continue

                dt_type: Optional[DataType] = get_instantiated_type(
                    annotation, optional=optional
                )
                if not dt_type:
                    unsupported_types[inp.name] = annotation
                    continue

                ins.append(
                    TInput(
                        tag=inp.name,
                        intype=dt_type,
                        default=default,
                        doc=InputDocumentation(paramdocs.get(inp.name)),
                    )
                )

            if missing_annotations:
                raise Exception(
                    f"The following types on the PythonTool '{self.id()}' were missing type annotations (REQUIRED): "
                    + ", ".join(missing_annotations)
                )

            if unsupported_types:
                raise Exception(
                    f"Unsupported types for inputs: "
                    + ", ".join(f"{k}: {v}" for k, v in unsupported_types.items())
                )
            self._cached_input_signature = ins

        return self._cached_input_signature

    def base_command(self):
        return ["python", self.script_name()]

    def script_name(self):
        return f"{self.id()}-script.py"

    def container(self):
        return "python:3.8.1"

    def prepared_script(self):
        import inspect

        nl = "\n"
        argkwargs = ", ".join(f"{t.id()}=args.{t.id()}" for t in self.inputs())

        codeblock_without_static = nl.join(
            dedent(inspect.getsource(self.code_block)).split(nl)[1:]
        )

        ins = self.tool_inputs()

        pt_decl = """\
class PythonTool:
    File = str
    Directory = str
"""

        type_annotation_declarations = nl.join(
            f"{k} = {v}"
            for k, v in {
                **{inp.intype.__class__.__name__: "str" for inp in ins},
                **{
                    t.__name__: NativeTypes.map_to_primitive(t.primitive()).__name__
                    for t in all_types
                },
                "Array": "List",
            }.items()
        )

        return f"""
import argparse, json, sys
from typing import Optional, List, Dict, Any
cli = argparse.ArgumentParser("Argument parser for Janis PythonTool")
{nl.join(self.generate_cli_binding_for_input(inp) for inp in ins)}

{type_annotation_declarations}
{pt_decl}


{codeblock_without_static}

try:
    args = cli.parse_args()
    result = code_block({argkwargs})
    print(json.dumps(result))
except e:
    print(str(e), file=sys.stderr)
    raise
"""

    @staticmethod
    def generate_cli_binding_for_input(inp: TInput):
        params = [f'"--{inp.id()}"']

        intype = None

        bintype = inp.intype
        required = not inp.intype.optional

        if isinstance(bintype, Array):
            params.append("nargs='+'")
            bintype = bintype.fundamental_type()

        if isinstance(bintype, Int):
            intype = "int"
        elif isinstance(bintype, Float):
            intype = "float"
        elif isinstance(bintype, (String, File, Directory)):
            intype = "str"
        elif isinstance(bintype, Boolean):
            params.append("action='store_true'")
            required = False

        if intype:
            params.append("type=" + intype)

        if required:
            params.append("required=True")

        if inp.doc and inp.doc.doc:
            escaped = inp.doc.doc.replace("'", "\\'").replace("\n", "\\n")
            params.append(f"help='{escaped}'")

        joined = ", ".join(params)
        return f"cli.add_argument({joined})"
